@@ -7,7 +7,7 @@ use v5.20;
 use strict;
 use warnings;
 
-our $VERSION = '0.03'; # TRIAL
+our $VERSION = '0.04'; # TRIAL
 
 use Scalar::Util 'reftype';
 use List::Util 1.45 'uniqstr';
@@ -22,6 +22,8 @@ our %EXPORT_TAGS = (
     utils     => [qw( install_constant_tag install_constant_func )],
 );
 
+our %HOOK;
+
 install_EXPORTS();
 
 sub _croak {
@@ -123,6 +125,11 @@ sub install_EXPORTS {
     my $package     = delete $options{package};
     my $install_all = delete $options{all};
 
+    # run hooks.
+    if ( defined( my $hooks = delete $HOOK{$package}{pre} ) ) {
+        $_->() for values $hooks->%*;
+    }
+
     my $EXPORT_TAGS = _EXPORT_TAGS( $package );
 
     if ( defined $export_tags ) {
@@ -179,6 +186,33 @@ sub install_EXPORTS {
 
 
 
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
 
 
 
@@ -194,13 +228,14 @@ sub install_EXPORTS {
 
 
 sub install_CONSTANTS {
-    my $caller = !defined reftype( $_[-1] ) ? pop : scalar caller;
+    my $package = !defined reftype( $_[-1] ) ? pop : scalar caller;
 
     for my $spec ( @_ ) {
         my $type = reftype( $spec );
 
         if ( 'HASH' eq $type ) {
-            install_constant_tag( $_, $spec->{$_}, $caller ) for keys $spec->%*;
+            install_constant_tag( $_, $spec->{$_}, $package )
+              for keys $spec->%*;
         }
 
         elsif ( 'ARRAY' eq $type ) {
@@ -212,7 +247,7 @@ sub install_CONSTANTS {
             while ( $idx ) {
                 my $hash = $spec->[ --$idx ];
                 my $id   = $spec->[ --$idx ];
-                install_constant_tag( $id, $hash, $caller );
+                install_constant_tag( $id, $hash, $package );
             }
         }
 
@@ -348,7 +383,8 @@ sub install_CONSTANTS {
 
 
 
-sub install_constant_tag ( $id, $constants, $caller = scalar caller ) {
+
+sub install_constant_tag ( $id, $constants, $package = scalar caller ) {
 
     my ( @names, @values );
     if ( reftype( $constants ) eq 'HASH' ) {
@@ -356,16 +392,11 @@ sub install_constant_tag ( $id, $constants, $caller = scalar caller ) {
         @values = values $constants->%*;
     }
     elsif ( reftype( $constants ) eq 'ARRAY' ) {
-        my $idx = $constants->@*;
-        while ( $idx ) {
-            # pulling things off in reverse order, so it's (value,
-            # name), not (name, value)
-            push @values, $constants->[ --$idx ];
-            push @names,  $constants->[ --$idx ];
+        my @copy = $constants->@*;
+        while ( my ( $name, $value ) = splice( @copy, 0, 2 ) ) {
+            push @names,  $name;
+            push @values, $value;
         }
-        # reverse order to keep in user specified order, as it was
-        # constructed in reverse order
-        @values    = reverse @values;
         $constants = { $constants->@* };
     }
     else {
@@ -373,7 +404,7 @@ sub install_constant_tag ( $id, $constants, $caller = scalar caller ) {
             '$constants argument should be either a hashref or an arrayref' );
     }
 
-    constant->import::into( $caller, $constants );
+    constant->import::into( $package, $constants );
 
     # caller may specify distinct tag and enumeration function names.
     my ( $tag, $fname )
@@ -381,16 +412,18 @@ sub install_constant_tag ( $id, $constants, $caller = scalar caller ) {
       ? ( $id->[0], $id->[1] )
       : ( lc( $id ), $id );
 
-    push( ( _EXPORT_TAGS( $caller )->{$tag} //= [] )->@*, @names );
 
-    # supplement previous enumerating function, if it exists
-    my $fqfn = "${caller}::${fname}";
-    if ( defined &{$fqfn} ) {
+    push( ( _EXPORT_TAGS( $package )->{$tag} //= [] )->@*, @names );
+
+    my $fqdn = join '::', $package, $fname;
+    $HOOK{$package}{pre}{$fname} //= sub {
         no strict 'refs';    ## no critic
-        @values = ( &$fqfn, @values );
-    }
-
-    install_constant_func( $fname, \@values, $caller );
+        _croak( "Error: attempt to redefine enumerating function $fqdn" )
+          if exists &{$fqdn};
+        my @values
+          = map { &{"${package}::${_}"} } _EXPORT_TAGS( $package )->{$tag}->@*;
+        install_constant_func( $fname, \@values, $package );
+    };
 }
 
 
@@ -458,7 +491,7 @@ __END__
 
 =pod
 
-=for :stopwords Diab Jerius Smithsonian Astrophysical Observatory
+=for :stopwords Diab Jerius Smithsonian Astrophysical Observatory mistyped
 
 =head1 NAME
 
@@ -466,7 +499,7 @@ CXC::Exporter::Util - Tagged Based Exporting
 
 =head1 VERSION
 
-version 0.03
+version 0.04
 
 =head1 SYNOPSIS
 
@@ -564,8 +597,8 @@ L<CXC::Exporter::Util> provides additional support for creating,
 organizing and installing constants via L</install_CONSTANTS>.
 Constants are created via Perl's L<constant> pragma.
 
-L</install_CONSTANTS> is passed one or more hashes containing sets of
-constants grouped by tags, e.g.:
+L</install_CONSTANTS> is passed sets of constants grouped by tags,
+e.g.:
 
   install_CONSTANTS( {
         DETECTORS => {
@@ -584,7 +617,8 @@ constants grouped by tags, e.g.:
    # install_CONSTANTS;
    install_EXPORTS;
 
-For each set an enumerating function is created which returns the set's values.
+For each set an enumerating function is created which returns the
+set's values.
 
 In the above example, constant functions C<ACIS>, C<HRC>, C<ALL>,
 C<NONE>, C<ANY>, tags C<detectors> and C<aggregates>, and constant
@@ -599,15 +633,23 @@ or iterating:
 
   say $_ for DETECTORS;
 
-If the constants are used later in the module for other purposes, constant definition
+C<install_CONSTANTS> may be called multiple times. If the constants
+are used later in the module for other purposes, constant definition
 should be done in a L<BEGIN> block:
 
-  our %CONSTANTS;
-  BEGIN { %CONSTANTS = (...);
-          install_CONSTANTS( \%CONSTANTS );
+  BEGIN {
+      install_CONSTANTS( {
+          CCD => {nCCDColumns  => 1024, minCCDColumn => 0,},
+      } );
   }
 
-C<install_CONSTANTS> may be called multiple times.
+  install_CONSTANTS( {
+      CCD => {
+          maxCCDColumn => minCCDColumn + nCCDColumns - 1,
+      } }
+  );
+
+  install_EXPORTS;
 
 For more complex situations, the lower level L</install_constant_tag>
 and L</install_constant_func> routines may be useful.
@@ -633,8 +675,8 @@ If the C<overwrite> option is true, the contents of C<%export_tags>
 will overwrite C<%EXPORT_TAGS> in C<$package>, otherwise
 C<%export_tags> is merged into C<%EXPORT_TAGS>.
 
-Note that overwriting will wipe out tags and symbols installed by a
-previous call to L</install_CONSTANTS>.
+Note that overwriting will remove the tags and symbols installed into
+C<%EXPORT_TAGS> by previous calls to L</install_CONSTANTS>.
 
 This defaults to false.
 
@@ -682,26 +724,53 @@ C<$package>.
 If not specified,  C<$package> defaults to the caller's package.
 
 The passed C<@specs> arguments are either hashrefs or arrayrefs and
-contain one or more constant set specifications.  A constant set
-specification consists of a unique identifier and a list of name-value
-pairs, specified either has a hash or an array.  For example,
+contain one or more set specifications.  A set specification
+consists of a unique identifier and a list of name-value pairs,
+specified either as a hash or an array.  For example,
 
   @spec = ( { $id1 => \%set1, $id2 => \@set2 },
             [ $id3 => \%set3, $id4 => \@set4 ],
           );
 
-The C<$id?> identifiers are used to create a export tag for the set as
-well as an enumerating function which returns the set's values.
-
+The identifier is used to create an export tag for the set, as
+well as to name an enumerating function which returns the set's values.
 The individual C<$id>, C<$set> pairs are passed to L<install_constant_tag>;
-see that function for more information.
+see that function for more information on how the identifiers are used.
 
-L</install_CONSTANTS> may be called multiple times, and may be used
-to updated existing sets of constants.
+A call to L<install_EXPORTS> I<must> be made after the last call to
+C<install_CONSTANTS> or
 
-Note that a call to L<install_EXPORTS> I<must> be made after the last
-call to C<install_CONSTANTS> or the constants won't be added to the
-exports.
+=over
+
+=item *
+
+The constants won't be added to the exports.
+
+=item *
+
+The enumerating functions won't be created.
+
+=back
+
+L</install_CONSTANTS> may be called more than once to add symbols to a tag,
+but don't split those calls across a call to L</install_EXPORTS>.
+
+In other words,
+
+  # DON"T DO THIS, IT'LL THROW
+  install_CONSTANTS( { Foo => { bar => 1 } } );
+  install_EXPORTS;
+  install_CONSTANTS( { Foo => { baz => 1 } } );
+  install_EXPORTS;
+
+  # DO THIS
+  install_CONSTANTS( { Foo => { bar => 1 } } );
+  install_CONSTANTS( { Foo => { baz => 1 } } );
+  install_EXPORTS;
+
+Each call to L</install_EXPORTS> installs the enumerating functions for
+sets modified since the last call to it, and each enumerating function
+can only be added once.
 
 =head2 install_constant_tag
 
@@ -731,26 +800,28 @@ L</install_constant_tag> will
 
 =item 1
 
-use Perl's L<constant> pragma to create a function named I<name> returning I<value>
-for each I<name>-I<value> pair in C<$constants>.
+use Perl's L<constant> pragma to create a function named I<name>
+returning I<value> for each I<name>-I<value> pair in C<$constants>.
 
-The functions are installed in C<$package> and their names added to
-the symbols in C<%EXPORT_TAGS> with export tag C<$tag>.
+The functions are installed in C<$package> and their names appended to
+the symbols in C<%EXPORT_TAGS> with export tag C<$tag>.  If C<$constants>
+is an arrayref they are appended in the ordered specified in the array,
+otherwise they are appended in random order.
 
 =item 2
 
-use Perl's L<constant> pragma to create an enumerating function named
-C<$fname> which returns a list of the I<values> in C<$constants>.  If
-C<$constants> is an arrayref, the values will be returned in the order
-specified.
+Add a hook so that the next time L</install_EXPORTS> is called, Perl's
+L<constant> pragma will be used to create an enumerating function
+named C<$fname> which returns a list of the I<values> of the constants
+associated with C<$tag>, in the order they were added to
+C<$EXPORT_TAGS{$tag}>.
 
-If the function C<$fname> already exists, a new function is created
-(named C<$fname>) which returns the values returned by the original
-function and the new constant values added by this call to
-L</install_constant_tag>.
+The enumerating function C<$fname> is added to the symbols in
+C<%EXPORT_TAGS> tagged with C<contant_funcs>.
 
-C<$fname> is added to the symbols in C<%EXPORT_TAGS> tagged
-with C<contant_funcs>.
+Just as you shouldn't interleave calls to L</install_CONSTANTS> for a
+single tag with calls to L</install_EXPORTS>, don't interleave calls
+to L</install_constant_tag> with calls to L</install_EXPORTS>.
 
 =back
 
@@ -759,6 +830,7 @@ For example, after
   $id = 'AGGREGATES';
   $constants = { ALL => 'all', NONE => 'none', ANY => 'any' };
   install_constant_tag( $id, $constants );
+  install_EXPORTS:
 
 =over
 
@@ -815,16 +887,14 @@ constant functions and an enumerating function called C<Animals>.
 C<install_constant_tag> uses L</install_constant_func> to create and install
 the constant functions which return the constant values.
 
-B<Caveat!> Because of how enumerating functions are created,
-specifying new values and a new enumerating function name for an
-existing tag will not cause the new values to be added to the previous
-enumerating function. For example,
+Because of when enumerating functions are created, all enumerating functions
+associated with a set will return all of the set's values, regardless of when
+the function was specified.  For example,
 
   install_constant_tag( 'TAG', { HIGH => 'high' }  );
   install_constant_tag( [ 'TAG', 'ATAG' ], { LOW => 'low' } );
 
-Will create a function C<TAG> which returns C<high> and a function
-C<ATAG> which returns C<low>.  Don't depend upon this behavior.
+will create functions C<TAG> and C<ATAG> which both return C<high>, C<low>.
 
 =head2 install_constant_func( $name, \@values, $caller )
 
@@ -900,6 +970,40 @@ and then
 
   say Sun | Mon;
 
+=item Using a constant in the exporting module
+
+When a constant is used in exporting module (to create another constant, for example),
+it's tempting to do something like this:
+
+  # DON"T DO THIS
+  %CCD = ( minCCDColumn => 0, nCCDColumns = 1024 );
+  $CCD{maxCCDColumn} = $CCD{minCCDColumn} + $CCD{nCCDColumns} - 1;
+  install_CONSTANTS( { CCD => \%CCD } );
+  install_EXPORTS;
+
+Not only is this noisy code, if the hash keys are mistyped, there's an
+error, which is exactly what constants are supposed to avoid.
+
+Instead, create an initial set of constants in a BEGIN block, which
+will make them available for the rest of the code:
+
+  BEGIN {
+      install_CONSTANTS( {
+          CCD => {nCCDColumns  => 1024, minCCDColumn => 0,},
+      } );
+  }
+
+  install_CONSTANTS( {
+      CCD => {
+          maxCCDColumn => minCCDColumn + nCCDColumns - 1,
+      } }
+  );
+
+  install_EXPORTS;
+
+A bit more verbose, but it uses the generated constant functions and
+avoids errors.
+
 =back
 
 =head1 SUPPORT
@@ -928,6 +1032,14 @@ Please see those modules/websites for more information related to this module.
 
 L<Exporter|Exporter>
 
+=item *
+
+L<Exporter::Tiny|Exporter::Tiny>
+
+=item *
+
+L<Exporter::Almighty|Exporter::Almighty>
+
 =back
 
 =head1 AUTHOR
